
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "transcoder-enumtypes.h"
#include "gsttranscoder.h"
#include "transcoder-prelude.h"
#include "gsttranscoder-signal-adapter.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gsttranscoder.h" */

GType
gst_transcoder_error_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_TRANSCODER_ERROR_FAILED), "GST_TRANSCODER_ERROR_FAILED", "failed" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstTranscoderError"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_transcoder_state_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_TRANSCODER_STATE_STOPPED), "GST_TRANSCODER_STATE_STOPPED", "stopped" },
    { C_ENUM(GST_TRANSCODER_STATE_PAUSED), "GST_TRANSCODER_STATE_PAUSED", "paused" },
    { C_ENUM(GST_TRANSCODER_STATE_PLAYING), "GST_TRANSCODER_STATE_PLAYING", "playing" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstTranscoderState"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_transcoder_message_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_TRANSCODER_MESSAGE_POSITION_UPDATED), "GST_TRANSCODER_MESSAGE_POSITION_UPDATED", "position-updated" },
    { C_ENUM(GST_TRANSCODER_MESSAGE_DURATION_CHANGED), "GST_TRANSCODER_MESSAGE_DURATION_CHANGED", "duration-changed" },
    { C_ENUM(GST_TRANSCODER_MESSAGE_STATE_CHANGED), "GST_TRANSCODER_MESSAGE_STATE_CHANGED", "state-changed" },
    { C_ENUM(GST_TRANSCODER_MESSAGE_DONE), "GST_TRANSCODER_MESSAGE_DONE", "done" },
    { C_ENUM(GST_TRANSCODER_MESSAGE_ERROR), "GST_TRANSCODER_MESSAGE_ERROR", "error" },
    { C_ENUM(GST_TRANSCODER_MESSAGE_WARNING), "GST_TRANSCODER_MESSAGE_WARNING", "warning" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstTranscoderMessage"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* Generated data ends here */

